#include "gtkpapertapezoom.h"
#include <gtkmm/actiongroup.h>
#include <gtkmm/table.h>
#include <gtkmm/label.h>
#include <gtkmm/frame.h>
#include <gtkmm/checkbutton.h>
#include <gtkmm/box.h>
#include <gtk/gtk.h>
#include <cairo.h>
#include <gtkmm/stock.h>
#include <iostream>
#include <cstdio>

using namespace Gtk;


/* { name, stock_id, label, acclerator, tooltip, callback } */
GtkActionEntry PaperTapeZoom::ui_action_entries[] = {
	{ "ZoomIn", GTK_STOCK_ZOOM_IN },
	{ "ZoomOut", GTK_STOCK_ZOOM_OUT },
	{ "Zoom100", GTK_STOCK_ZOOM_100 },
	{ "Zoom50", NULL, "Zoom to 50%" },
	{ "Zoom200", NULL, "Zoom to 200%" },
	{ "ZoomWidth", GTK_STOCK_ZOOM_FIT, "Fit to screen", NULL, "Zoom that the papertape width fills the screen" },
	{ "ZoomLength", GTK_STOCK_ZOOM_FIT, "Fit all to screen ", NULL, "Zoom that the complete papertape fills in the screen" },
	{ "ZoomDialog", GTK_STOCK_PROPERTIES, "Show dialog...", "<ctrl>z", "Set Zoom level and input own Affine Translation Matrix" },

	{ "Rotate90CW", GTK_STOCK_ORIENTATION_LANDSCAPE, "Rotate clockwise" },
	{ "Rotate90CCW", GTK_STOCK_ORIENTATION_REVERSE_LANDSCAPE, "Rotate anticlockwise" },
	{ "FlipHoriz", GTK_STOCK_ORIENTATION_PORTRAIT, "Flip horizontally" },
	{ "FlipVert", GTK_STOCK_ORIENTATION_REVERSE_PORTRAIT, "Flip verticall" }
};

PaperTapeZoom::PaperTapeZoom(double default_rotation, double default_zoom_factor, PaperTapeFlip default_flip) {
	// init values
	this->rotation = default_rotation;
	this->flip = default_flip;
	this->zoom_factor = default_zoom_factor;
	this->manually = false;
	set_default(default_rotation, default_zoom_factor, default_flip);
}

void PaperTapeZoom::show_dialog() {
	if(!matrix_dialog) {
		// set up the matrix_dialog.
		matrix_dialog = new Dialog("Set papertape view");
		// Frame.
		Frame* matrix_frame = manage(new Frame());
		CheckButton* matrix_manually = manage(new CheckButton("Set Affine Translation Matrix manually"));
		matrix_manually->set_active(this->manually);
		// connect matrix_manually button
		matrix_manually->signal_toggled().connect(
			sigc::bind(sigc::mem_fun(*this, &PaperTapeZoom::on_toggle_manually), matrix_manually)
		);
		matrix_frame->set_label_widget(*matrix_manually);
		
		Table* table = manage(new Table(2, 6));
		Label* l;
		l = manage(new Label("x__n = "));
		table->attach(*l, 0, 1, 0, 1);
		matrix_entries[&xx] = manage(new SpinButton());
		table->attach(*matrix_entries[&xx], 1, 2, 0, 1);
		l = manage(new Label("* x + "));
		table->attach(*l, 2, 3, 0, 1);
		matrix_entries[&xy] = manage(new SpinButton());
		table->attach(*matrix_entries[&xy], 3, 4, 0, 1);
		l = manage(new Label("* y + "));
		table->attach(*l, 4, 5, 0, 1);
		matrix_entries[&x0] = manage(new SpinButton());
		table->attach(*matrix_entries[&x0], 5, 6, 0, 1);
		l = manage(new Label("y__n = "));
		table->attach(*l, 0, 1, 1, 2);
		matrix_entries[&yx] = manage(new SpinButton());
		table->attach(*matrix_entries[&yx], 1, 2, 1, 2);
		l = manage(new Label("* x + "));
		table->attach(*l, 2, 3, 1, 2);
		matrix_entries[&yy] = manage(new SpinButton());
		table->attach(*matrix_entries[&yy], 3, 4, 1, 2);
		l = manage(new Label("* y + "));
		table->attach(*l, 4, 5, 1, 2);
		matrix_entries[&y0] = manage(new SpinButton());
		table->attach(*matrix_entries[&y0], 5, 6, 1, 2);
		
		// connect signals
		for(std::map<double*, SpinButton*>::iterator i = matrix_entries.begin();
		    i != matrix_entries.end(); i++) {
			i->second->set_sensitive(this->manually);
			i->second->set_range(-G_MAXDOUBLE, G_MAXDOUBLE); // ;-)
			i->second->set_increments(5.0, 10.0);
			i->second->set_value( * (i->first) );
			i->second->signal_value_changed().connect(
				sigc::bind(
					sigc::mem_fun(*this, &PaperTapeZoom::on_change_field),
					i->first
				)
			);
		} // for entries

		
		matrix_frame->add(*table);
		matrix_dialog->get_vbox()->add(*matrix_frame);
		
		// add buttons
		//matrix_dialog->add_button(GTK_STOCK_UNDO, 1);
		Button* reset = manage(new Button(Stock::UNDO));
		matrix_dialog->get_action_area()->add(*reset);
		matrix_dialog->get_action_area()->set_child_secondary(*reset);
		reset->signal_clicked().connect(sigc::mem_fun(*this, &PaperTapeZoom::reset));

		//matrix_dialog->add_button(Stock::APPLY, 2);
		// directly apply! :-)
		
		// we could use this, but we could not conncet own signals, so:
		//matrix_dialog->add_button(Stock::CLOSE, 3);
		Button *close = manage(new Button(Stock::CLOSE));
		matrix_dialog->get_action_area()->add(*close);
		close->signal_clicked().connect(sigc::mem_fun(*matrix_dialog, &Window::hide));
		
		// update the entry things whenever the zoom is changed
		this->signal_changed().connect(sigc::mem_fun(*this, &PaperTapeZoom::update_dialog_on_zoom_changes));
		
		//matrix_dialog->add(*main_box);
		matrix_dialog->set_focus(*close);
		matrix_dialog->show_all();
	} else {
		// Window already created: Show it and/or bring it to front
		matrix_dialog->show();
		matrix_dialog->grab_focus(); // try it...
		matrix_dialog->raise(); // bring it to front
	}
}

void PaperTapeZoom::update_dialog_on_zoom_changes() {
	for(std::map<double*, SpinButton*>::iterator i = matrix_entries.begin();
	    i != matrix_entries.end(); i++) {
		i->second->set_value( * (i->first) );
	} // for entries
}

void PaperTapeZoom::reset() {
	if(this->is_controlled_manually()) {
		*this = default_matrix; /*xx = default_matrix.xx; xy = default_matrix.xy; x0 = default_matrix.x0;
		yx = default_matrix.yx; yy = default_matrix.yy; y0 = default_matrix.y0;*/
		zoom_changed.emit();
	} else {
		zoom_factor = 1.0;
		rotation = 0.0;
		calc_matrix();
	}
}

void PaperTapeZoom::on_toggle_manually(CheckButton* manually_toggle) {
	this->manually = manually_toggle->get_active();
	// turn input widgets on/off
	for(std::map<double*, SpinButton*>::iterator i = matrix_entries.begin();
	    i != matrix_entries.end(); i++) {
		i->second->set_sensitive(manually);
	}
	// + turn other widgets on/off, like action widgets
}

void PaperTapeZoom::on_change_field(double* field) {
	// only for internal use
	if(!field) {
		std::cout << "Bad NUll Pointer at on_change_field.\n";
		return;
	}
	if(matrix_entries[field]) {
		std::cout << "UPDATE FIELD\n";
		*field = matrix_entries[field]->get_value();
		zoom_changed.emit();
	} else
		std::cout << "BAD: FIELD NOT SUPPORTED (PaperTapeZoom::change_field)\n";
}

void PaperTapeZoom::calc_matrix() {
	// take the buffer things and calculate a new matrix, FROM
	// DEFAULT MATRIX.
	if(!this->is_controlled_manually()) {
		*this = default_matrix;
		cairo_matrix_scale(this, zoom_factor, zoom_factor);
		cairo_matrix_rotate(this, rotation);

		zoom_changed.emit();
	}
}

void PaperTapeZoom::set_default(double rotation, double zoom_factor, PaperTapeFlip flip) {
	// change default matrix.
	cairo_matrix_init_identity(&this->default_matrix);
	cairo_matrix_scale(&this->default_matrix, zoom_factor, zoom_factor);
	cairo_matrix_rotate(&this->default_matrix, rotation);
	// flip?
}

void PaperTapeZoom::set_rotate(double radians, bool relative) {
	if(!is_controlled_manually()) {
		this->rotation = relative ? (this->rotation + radians) : radians;
		this->calc_matrix();
	} else {
		// change user matrix
		cairo_matrix_rotate(this, radians);
		zoom_changed.emit();
	}
}

void PaperTapeZoom::set_zoom(double zoom_factor, bool relative) {
	if(!is_controlled_manually()) {
		this->zoom_factor = relative ? (this->zoom_factor * zoom_factor) : zoom_factor;
		this->calc_matrix();
	} else {
		// change user matrix
		cairo_matrix_scale(this, zoom_factor, zoom_factor);
		zoom_changed.emit();
	}
}

void PaperTapeZoom::set_flip(PaperTapeFlip flip, bool relative) {
	// not implemented yet
}

Glib::RefPtr<ActionGroup> PaperTapeZoom::get_action_group() {
	Glib::RefPtr<ActionGroup> group = ActionGroup::create("GtkPaperTapeZoom");
	
	gtk_action_group_add_actions(
			group->gobj(),
			ui_action_entries,
			G_N_ELEMENTS(ui_action_entries),
			NULL);

	group->signal_post_activate().connect(sigc::mem_fun(*this, &PaperTapeZoom::transform_action));

	return group;
}

void PaperTapeZoom::transform_action(const Glib::RefPtr<Action>& action) {
	Glib::ustring name = action->get_name();
	
	if(name == "ZoomIn") {
		set_zoom(1.1, true);
	} else if(name == "ZoomOut") {
		set_zoom(0.9, true);
	} else if(name == "Zoom100") {
		set_zoom(1.0, false);
	} else if(name == "Zoom50") {
		set_zoom(0.5, false);
	} else if(name == "Zoom200") {
		set_zoom(2.0, false);
	} else if(name == "ZoomWidth") {
		//signal_set_auto_zoom.emit(true);
		
		//set_auto_zoom(ZOOM_LOCHSTREIFEN_WIDTH);
		// get width of lochstreifen (well, "1")
		// get height of view Viewport
		// calc length
		// add signal to fetch size changements
	} else if(name == "ZoomLength") {
		//set_auto_zoom(ZOOM_LOCHSTREIFEN_LENGTH);
	} else if(name == "ZoomDialog") {
		show_dialog();
	} else if(name == "Rotate90CW") {
		set_rotate(G_PI, true);
	} else if(name == "Rotate90CCW") {
		set_rotate(-G_PI, true);
	} else if(name == "FlipHoriz") {
		//set_flip();
	} else if(name == "FlipVert") {
		// same here...
	} else {
		std::cout << "PaperTapeZoom: Action " << name << " not supported!\n";
	}
}

void PaperTapeZoom::set_auto_zoom(PaperTapeView& view, PaperTapeAutoZoomType type) {
	this->auto_zoom_enabled = true;

	// connect resize signal
	auto_zoom_resize_con = view.signal_configure_event().connect(
		sigc::mem_fun(*this, &PaperTapeZoom::on_auto_zoom_resize)
	);

	
	
}

void PaperTapeZoom::disable_auto_zoom() {
	if(auto_zoom_enabled) {
		auto_zoom_resize_con.disconnect();
		auto_zoom_enabled = false;
	}
}

bool PaperTapeZoom::on_auto_zoom_resize(GdkEventConfigure* event) {
	// depending on orientation, use event->width and event->height
	
	// We could use LOCHSTREIFEN logic here!
	return true;
}


/*
void PaperTapeView::transform(PaperTapeTransformation how) {
	switch(how) {
		case ZOOM_IN:
		case ZOOM_OUT:
		{
			//current_width_scaling += 10 * ((how == ZOOM_IN) ? 1 : -1);
			double f((how == ZOOM_IN) ? 1.1 : (0.9));
			
			std::cout << "View: scaling up/down: " << f << "\n";

			cairo_matrix_scale(&lochstreifen->matrix, f, f);
			lochstreifen_check_matrix(lochstreifen);
			zoom_changed.emit();
			break;
		}
		case ZOOM_REAL:
		{
			// get the DPI, our factor
			double f = Gdk::Screen::get_default().get_resolution();
			cairo_matrix_init_scale(&lochstreifen->matrix, f, f);
			lochstreifen_check_matrix(lochstreifen);
			zoom_changed.emit();
		}
		case SET_ZOOM:
		{
			// display the matrix enter dialog
		}
	}
}
*/
